from kubernetes import client, config
from kubernetes.client import ApiClient
from rich.console import Console

from pbsrollout.k8s_utils import import_kube_config


def check_cluster_ready(name: str, path: str, api_client: ApiClient) -> bool:
    console = Console()
    print(f'\tChecking cluster {name}')
    daemonset_check = check_only_one_daemonset(path)
    if daemonset_check:
        console.print("\t\tonly one daemonset:................. [bold green]SUCCESS")
    else:
        console.print("\t\tonly one daemonset:................. [bold red]FAILED")
        return False

    return True


def check_only_one_daemonset(path: str) -> bool:
    config.load_kube_config(config_file=import_kube_config(path + '/.k8s-assets/kubeconfig'))

    v1 = client.AppsV1Api()
    daemonsets = v1.list_namespaced_daemon_set(namespace='api')
    if len(daemonsets.items) == 0:
        print('\t\terror: no daemonset found')
        return False

    pbs_daemonsets = []
    for d in daemonsets.items:
        name = d.metadata.name
        if name.startswith('pbs-prod'):
            pbs_daemonsets.append(name)
    if len(pbs_daemonsets) == 1:
        return True
    else:
        print(f'\t\terror: more than 1 daemonset found: {pbs_daemonsets}')
        return False



from typing import Optional

import pync
from rich.console import Console
from rich.syntax import Syntax

Error = Optional[str]


def notify(message, title, sound=True):
    if sound:
        pync.notify(message, title=title, sound='Ping')
    else:
        pync.notify(message, title=title)


def print_error_body(err: str):
    syntax = Syntax(err, "text", theme="paraiso-dark", line_numbers=False)
    Console().print(syntax)


def clean_stderr(s: str):
    if s is None:
        return ""
    s = s.split("\n")
    o = []
    for l in s:
        if "direnv" in l:
            continue
        o.append(l)

    if len(o) == 0:
        return ""
    return "\n".join(o)



import os
from unittest import TestCase

from kubernetes import config

from pbsrollout.k8s_utils import import_kube_config
from pbsrollout.process_cluster import remove_old_pods


class Test(TestCase):
    def test_remove_old_pods(self):
        path = os.environ['GOPATH'] + '/src/github.com/publica-project/prebid-kube-manifests/aws.us-east-1-eks-23-v3'
        api_client = config.new_client_from_config(config_file=import_kube_config(path + '/.k8s-assets/kubeconfig'))

        err = remove_old_pods('tesk_k8s', path, api_client)
        self.assertEqual('', err)



import os
from unittest import TestCase

from kubernetes import config

from pbsrollout.k8s_utils import import_kube_config, get_pods_for_daemonsets, get_all_pbs_daemonset, get_services


class Test(TestCase):
    def test_get_pods_for_daemonset(self):
        path = os.environ['GOPATH'] + '/src/github.com/publica-project/prebid-kube-manifests/aws.eu-west-1-eks-16-v3'
        api_client = config.new_client_from_config(config_file=import_kube_config(path + '/.k8s-assets/kubeconfig'))

        pods = get_pods_for_daemonsets(api_client, [802, 803])
        print(pods)

    def test_get_all_pbs_daemonset(self):
        path = os.environ['GOPATH'] + '/src/github.com/publica-project/prebid-kube-manifests/aws.eu-west-1-eks-16-v3'
        api_client = config.new_client_from_config(config_file=import_kube_config(path + '/.k8s-assets/kubeconfig'))

        get_all_pbs_daemonset(api_client)

    def test_get_services(self):
        path = os.environ['GOPATH'] + '/src/github.com/publica-project/prebid-kube-manifests/aws.eu-west-1-eks-16-v3'
        api_client = config.new_client_from_config(config_file=import_kube_config(path + '/.k8s-assets/kubeconfig'))

        get_services(api_client, 'pbs-prod')



import subprocess
from enum import Enum

from kubernetes.client import ApiClient
from rich.console import Console
from rich.padding import Padding
from rich.table import Table
from shellpython.helpers import Dir
from simple_term_menu import TerminalMenu

from pbsrollout.k8s_utils import get_all_pbs_daemonset, delete_pbs_daemonset, get_all_pblog_daemonset, get_services
from pbsrollout.utils import Error, print_error_body, clean_stderr


def launch_pods(name: str, path: str, api_client: ApiClient) -> bool:
    with Dir(path):
        print(f'\tLaunching new pods for cluster {name}')
        # run make clean deploy-new-pbs-prod
        cmd = "direnv exec . make clean deploy-new-pbs-prod"
        ret = subprocess.run(cmd.split(" "), stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, timeout=30)
        ret.stderr = clean_stderr(ret.stderr)
        if ret.returncode != 0 or ret.stderr != "":
            print_error_body(f"ERROR:\n\nstdout:\n{ret.stdout}\nstderr:\n{ret.stderr}")
            return False

    return True


def format_pods_status(desired: int, value: int) -> str:
    if desired != value:
        return f"[red bold] {value}"
    return f"[bold green] {value}"


def gen_row_daemonset(table, name, daemonset):
    desired = daemonset.status.desired_number_scheduled
    table.add_row(
        name,
        format_pods_status(desired, daemonset.status.desired_number_scheduled),  # Desired
        format_pods_status(desired, daemonset.status.current_number_scheduled),  # Current
        format_pods_status(desired, daemonset.status.updated_number_scheduled),  # Up-To_Date
        format_pods_status(desired, daemonset.status.number_ready),  # Ready
        format_pods_status(desired, daemonset.status.number_available),  # Available
    )


class Status(Enum):
    SUCCESS = 1
    WAIT = 2
    FAIL = 3


def display_daemonsets_informations(name: str, api_client: ApiClient) -> bool:
    console = Console()
    # 1. get all services
    # pbs
    pbs_daemonsets = get_all_pbs_daemonset(api_client)
    pbs_daemonsets_names = [d.metadata.name for d in pbs_daemonsets]
    if len(pbs_daemonsets_names) != 2:
        console.print(f"\t\t[bold red]error: should have 2 pbs daemonsets, got: {pbs_daemonsets_names}")
        return False

    pbs_daemonsets_names = [p.replace("pbs-prod-", "") for p in pbs_daemonsets_names]

    # pblog
    pblog_daemonsets = get_all_pblog_daemonset(api_client)
    pblog_daemonsets_names = [d.metadata.name for d in pblog_daemonsets]
    if len(pblog_daemonsets_names) != 2:
        console.print(f"\t\t[bold red]error: should have 2 pblog daemonsets, got: {pblog_daemonsets_names}")
        return False

    pblog_daemonsets_names = [p.replace("pblog-prod-", "") for p in pblog_daemonsets_names]

    # current service
    services = get_services(api_client, 'pbs-prod')
    services_names = [d.metadata.name for d in services]
    if services is None or len(services) != 1:
        console.print(f"\t\t[bold red]error: should have 1 service, got: {services_names}")
        return False

    # check that service is same as old pods and that next service will be same as new pods
    old_service_version = int(services[0].metadata.labels['version'])
    if old_service_version != int(pbs_daemonsets_names[0].replace("v", "")):
        console.print(f"\t\t[bold red]error: expected old pbs to have service version ({old_service_version}), got: {pbs_daemonsets_names[0]}")
        return False
    if old_service_version + 1 != int(pbs_daemonsets_names[1].replace("v", "")):
        console.print(f"\t\t[bold red]error: expected new pbs to have service version ({old_service_version + 1}), got: {pbs_daemonsets_names[1]}")
        return False

    console.print(f'\t\tswitching service {pbs_daemonsets_names[0]}->{pbs_daemonsets_names[1]}')
    console.print(f'\t\t[grey66]current service: {old_service_version}')

    # 2. Display pods infos for each daemonset
    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Daemonset", style="bold", width=12)
    table.add_column("Desired", style="bold")
    table.add_column("Current")
    table.add_column("Up-To-Date")
    table.add_column("Ready")
    table.add_column("Available")

    # pbs
    gen_row_daemonset(table, f"[grey66]pbs-{pbs_daemonsets_names[0]}", pbs_daemonsets[0])
    gen_row_daemonset(table, f"pbs-{pbs_daemonsets_names[1]}", pbs_daemonsets[1])
    # pblog
    gen_row_daemonset(table, f"[grey66]pblog-{pblog_daemonsets_names[0]}", pblog_daemonsets[0])
    gen_row_daemonset(table, f"pblog-{pblog_daemonsets_names[1]}", pblog_daemonsets[1])

    padded_table = Padding(table, (0, 0, 0, 16))
    console.print(padded_table)

    return True


def cutover_service(name: str, path: str, api_client: ApiClient) -> Status:
    console = Console()
    console.print(f'\tCutting over {name} to new service')
    with Dir(path):
        ok = display_daemonsets_informations(name, api_client)
        if not ok:
            return Status.FAIL

        terminal_menu = TerminalMenu(["[y] yes", "[n] no, let's wait"], title=f"Should we cutover?",
                                     raise_error_on_interrupt=True)
        menu_entry_index = terminal_menu.show()
        if menu_entry_index == 1:
            return Status.WAIT

        # do the cutover
        # make next-pbs-prod-service
        cmd = "direnv exec . make deploy-new-pbs-prod-service"
        ret = subprocess.run(cmd.split(" "), stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, timeout=30, input='y')
        ret.stderr = clean_stderr(ret.stderr)
        if ret.returncode != 0 or ret.stderr != "":
            print_error_body(f"ERROR:\n\nstdout:\n{ret.stdout}\nstderr:\n{ret.stderr}")
            return Status.FAIL
    return Status.SUCCESS


# return error
def remove_old_pods(name: str, path: str, api_client: ApiClient) -> Error:
    console = Console()
    with Dir(path):
        console.print(f'\tremoving old pods/daemonset for cluster {name}')

        # 1. Get pbs service selected version
        services = get_services(api_client, 'pbs-prod')
        if services is None or len(services) != 1:
            return "should have 1 service, got: {services_names}"
        selected_version = int(services[0].metadata.labels["version"])
        console.print(f'\t\t[grey66]currently selected version: {selected_version}')

        # 2. Get all daemonsets
        pbs_daemonsets = get_all_pbs_daemonset(api_client)
        pbs_daemonsets_names = [d.metadata.name for d in pbs_daemonsets]
        console.print(f'\t\tdaemonsets found: {pbs_daemonsets_names}')
        if len(pbs_daemonsets_names) != 2:
            return f'2 pbs daemonset expected (old/new). found: {pbs_daemonsets_names}'

        version1 = int(pbs_daemonsets_names[0].replace('pbs-prod-v', ''))
        version2 = int(pbs_daemonsets_names[1].replace('pbs-prod-v', ''))

        # remove the oldest version THAT IS NOT SELECTED
        if version1 == selected_version and version2 > selected_version:
            return f'selected version {selected_version} should be greater than the other version: {version2}'
        if version2 == selected_version and version1 > selected_version:
            return f'selected version {selected_version} should be greater than the other version: {version1}'
        if selected_version != version1 and selected_version != version2:
            return f'selected version {selected_version} doesnt match the 2 existing versions: {[version1, version2]} from {pbs_daemonsets_names}'

        # remove the oldest version
        version_to_remove = version1 if selected_version == version2 else version2
        console.print(f'\t\tremoving old version {version_to_remove}')
        delete_pbs_daemonset(api_client, version_to_remove)

        return None



import tempfile
from typing import List

from kubernetes import client
from kubernetes.client import ApiClient, V1DaemonSet, V1Service

data_to_append = """      env:
        - name: AWS_PROFILE
          value: "ctv_engineer"
"""


# Add aws env vars to the kubeconfig file. save it in a temp place and return the new path
def import_kube_config(path: str) -> str:
    with open(path) as fp:
        lines = fp.readlines()[:-1]
        for d in data_to_append.split('\n'):
            lines.append(d + '\n')
        with tempfile.NamedTemporaryFile(delete=False, mode='w') as tmp:
            tmp.writelines(lines)
            tmp.close()
            return tmp.name


# return ordered list of daemonset
def get_all_daemonset_filtered_by_name(api: ApiClient, startwith: str):
    v1 = client.AppsV1Api(api_client=api)

    daemonsets = v1.list_namespaced_daemon_set(namespace='api')
    if len(daemonsets.items) == 0:
        return []

    pbs_daemonsets = []
    for d in daemonsets.items:
        name = d.metadata.name
        if name.startswith(startwith):
            pbs_daemonsets.append(d)
    return sorted(pbs_daemonsets, key=lambda x: x.metadata.name)


def get_all_pbs_daemonset(api: ApiClient) -> List[V1DaemonSet]:
    return get_all_daemonset_filtered_by_name(api, 'pbs-prod')


def get_all_pblog_daemonset(api: ApiClient) -> List[V1DaemonSet]:
    return get_all_daemonset_filtered_by_name(api, 'pblog-prod')


def delete_pbs_daemonset(api: ApiClient, version: int):
    v1 = client.AppsV1Api(api_client=api)
    # delete pbs daemonset
    v1.delete_namespaced_daemon_set(namespace='api', name=f'pbs-prod-v{version}')
    # delete pblog daemonset
    v1.delete_namespaced_daemon_set(namespace='api', name=f'pblog-prod-v{version}')


def get_pods_for_daemonsets(api: ApiClient, daemonset_versions: List[int]):
    core = client.CoreV1Api(api_client=api)
    pods = core.list_namespaced_pod(namespace='api')
    output = {}
    for d in daemonset_versions:
        output[d] = []
    wanted = [f'pbs-prod-v{v}' for v in daemonset_versions]
    for p in pods.items:
        name = p.metadata.name
        for i, w in enumerate(wanted):
            if name.startswith(w):
                output[daemonset_versions[i]].append(p)
                break
    return output


def get_services(api: ApiClient, startwith: str) -> List[V1Service]:
    core = client.CoreV1Api(api_client=api)
    services = core.list_namespaced_service(namespace='api')
    if startwith == "":
        return services
    output = []
    for s in services.items:
        if s.metadata.name.startswith(startwith):
            output.append(s)
    return sorted(output, key=lambda x: x.metadata.name)



import os
import sys
import time
from subprocess import TimeoutExpired
from typing import Dict

from kubernetes import config
from kubernetes.client import ApiException, ApiClient
from rich.columns import Columns
from rich.console import Console
from rich.markdown import Markdown
from shellpython.helpers import Dir
from simple_term_menu import TerminalMenu

from pbsrollout.check_cluster_ready import check_cluster_ready
from pbsrollout.k8s_utils import import_kube_config
from pbsrollout.process_cluster import cutover_service, launch_pods, remove_old_pods, Status
from pbsrollout.utils import notify

CLUSTERS = [
    "aws.us-east-1-eks-23-v3",
    "aws.us-east-1-eks-24-v3",
    "aws.us-east-1-eks-25-v3",
    "aws.us-east-1-eks-26-v3",
    "aws.eu-west-1-eks-16-v3",
    "aws.ap-southeast-2-eks-33-v1",
    "aws.ap-southeast-2-eks-34-v1",
]

DD_BOARD_URL = "https://app.datadoghq.com/dashboard/e23-wak-w5b/prebid?live=true"


def print_help(console):
    mk = """
    Supported features:
    - Check that no other version are present
    - Launch pods + cutover
    - Check the service version + daemonset version
    - Clean pods

    Unsupported features (for now):
    - Rollback + remove old pods
    """

    md = Markdown(mk)
    console.print(md)


def print_welcome(console):
    console.print(
        "[blink underline italic dark_orange3]Better[/blink underline italic dark_orange3] [bold medium_purple3]Rollout[/bold medium_purple3]\n\n",
        justify="center")
    regions = {
        'US': [],
        'EU': [],
        'AUS': [],
    }
    for c in CLUSTERS:
        if 'eu-' in c:
            regions['EU'].append(c)
        elif 'us-' in c:
            regions['US'].append(c)
        elif 'ap-' in c:
            regions['AUS'].append(c)
        else:
            raise RuntimeError(f'unknown region for cluster: {c}')

    user_renderables = [f"[bold]{r}:[/bold]\n" + "\n".join([f"- [slate_blue1]{c}[/slate_blue1]" for c in regions[r]])
                        for r in regions]
    console.print(Columns(user_renderables), justify="center")


def get_cluster_name(c):
    return c.replace('aws.us-east-1-', '').replace('aws.eu-west-1-', '').replace('aws.ap-southeast-2-', '')


def update_prebid_tag(path: str, tag: str):
    path += "/namespace/api/pbs.prod.tag"
    print(f"Updating prebid tag to {tag} in file: {path}")
    with open(path, "w") as f:
        f.write(tag)


def ask_for_tag() -> str:
    tag = input("enter the prebid tag here (you can find it on circleci): ")
    terminal_menu = TerminalMenu(["[y] yes", "[n] no"], title=f"Is this your tag? --{tag}--",
                                 raise_error_on_interrupt=True)
    menu_entry_index = terminal_menu.show()
    if menu_entry_index == 0:
        return tag
    return ask_for_tag()


def ask_to_open_datadog():
    terminal_menu = TerminalMenu(["[y] yes", "[n] I will open it myself!"], title=f"Please check datadog! Do you want me to open it for you?\n{DD_BOARD_URL}",
                                 raise_error_on_interrupt=True)
    menu_entry_index = terminal_menu.show()
    if menu_entry_index == 0:
        os.system(f"open '{DD_BOARD_URL}'")


def ask_if_everything_look_fine_after_all_cutover() -> bool:
    terminal_menu = TerminalMenu(["[y] yes", "[n] no"], title=f"Is everything fine? (check datadog)\n{DD_BOARD_URL}",
                                 raise_error_on_interrupt=True)
    menu_entry_index = terminal_menu.show()
    return menu_entry_index == 0


def progress_sleep(console, t: int):
    with console.status(f"[bold green]Sleeping {t}seconds...") as status:
        for i in range(t):
            status.update(status=f"[bold green]Sleeping {t - i}seconds...")
            time.sleep(1)
    # for _ in tqdm(range(t), leave=False, desc=f"sleeping {t}s"):
    #     time.sleep(1)


def rollback(gopath: str, prebid_k8s_path, api_client: ApiClient):
    raise RuntimeError("NOT IMPLEMENTED :( Do a manual rollback, sorry :sad:")
    print("Step [X]: Rolling back the pods")
    print("Step [X-1]: Rollback to previous service")
    # TODO

    print("Step [X-2]: Remove new pods")
    # TODO


def _main():
    console = Console()

    if 'GOPATH' not in os.environ or os.environ['GOPATH'] == '':
        print('GOPATH should be set in your environ! Contact a publica engineer for help')
        return 1
    gopath = os.environ['GOPATH']

    # 1. CD Inside the prebid kube manifest directory
    prebid_k8s_path = gopath + '/src/github.com/publica-project/prebid-kube-manifests'
    with Dir(prebid_k8s_path):
        console.print("[bold]Step 1: updating prebid tag")
        tag = ask_for_tag()
        if tag == "":
            print("error: tag must not be empty")
            return

        update_prebid_tag(prebid_k8s_path, tag)

        # dev step: assign k8s clients to clusters map
        clusters_to_k8s_client = {}  # type: Dict[str, ApiClient]
        for c in CLUSTERS:
            clusters_to_k8s_client[c] = config.new_client_from_config(
                config_file=import_kube_config(prebid_k8s_path + '/' + c + '/.k8s-assets/kubeconfig'))

        console.print("\n[bold]Step 2: Checking that the cluster is ready")
        for c in CLUSTERS:
            success = check_cluster_ready(get_cluster_name(c), prebid_k8s_path + '/' + c, clusters_to_k8s_client[c])
            if not success:
                raise RuntimeError(f"cluster {get_cluster_name(c)} is not ready")

        console.print("\n[bold]Step 3: Launching the new pods")
        for c in CLUSTERS:
            success = launch_pods(get_cluster_name(c), prebid_k8s_path + '/' + c, clusters_to_k8s_client[c])
            if not success:
                raise RuntimeError(f"failed to launch pods for cluster {get_cluster_name(c)}")
            progress_sleep(console, 60)

        ask_to_open_datadog()

        console.print("\n[bold]Step 4: Cutting over to the new service")
        for c in CLUSTERS:
            while True:
                status = cutover_service(get_cluster_name(c), prebid_k8s_path + '/' + c, clusters_to_k8s_client[c])
                if status == Status.WAIT:
                    progress_sleep(console, 30)
                    continue
                if status == Status.SUCCESS:
                    break
                if status == Status.FAIL:
                    raise RuntimeError(f"failed to cutover the service for cluster {get_cluster_name(c)}")

        is_everything_fine_after_cutover = ask_if_everything_look_fine_after_all_cutover()
        if not is_everything_fine_after_cutover:
            rollback(gopath, prebid_k8s_path, clusters_to_k8s_client[c])
            return 1

        console.print("\n[bold]Step 5: Remove old pods")
        for c in CLUSTERS:
            err = remove_old_pods(get_cluster_name(c), prebid_k8s_path + '/' + c, clusters_to_k8s_client[c])
            if err is not None:
                raise RuntimeError(f'failed to remove old pods for cluster: {get_cluster_name(c)}\n{err}')

        notify("Rollout prebid successfull!", "PBS Rollout")

        console.print("\n[bold green]Done! Don't forget to add the file to git and commit it!")
        console.print(
            f"[grey66]git add namespace/api/pbs.prod.tag; git commit -m 'updating pbs to tag: {tag} with better_rollout'; git push")
        return 0


def main():
    console = Console()
    print_welcome(console)
    if 'help' in sys.argv:
        print_help(console)
    print("\n\n")
    try:
        _main()
    except Exception as e:
        notify("Error during Rollout prebid!", "PBS Rollout", sound=True)
        if isinstance(e, ApiException) and e.reason == 'Unauthorized':
            console.print("[bold red]\nerror: please run `zaws_ctv_engineer` to login")
            return 1
        elif isinstance(e, RuntimeError):
            console.print(f"[bold red]\nError: {e}")
        elif isinstance(e, TimeoutExpired):
            console.print(f"[bold red]\nerror: {e}")
        else:
            console.print_exception()
            console.print('[orange3]please send this to Dimitri Wyzlic on Slack!')
            return 1
    except KeyboardInterrupt:
        console.print("\n\n[bold red]Interrupted by user (keyboard -> CTRL-C)")
    return 0


if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pbsrollout']

package_data = \
{'': ['*']}

install_requires = \
['kubernetes>=24.2.0,<25.0.0',
 'pync>=2.0.3,<3.0.0',
 'rich>=12.5.1,<13.0.0',
 'shellpy>=0.5.1,<0.6.0',
 'simple-term-menu>=1.5.0,<2.0.0']

entry_points = \
{'console_scripts': ['pbsrollout = pbsrollout.main:main']}

setup_kwargs = {
    'name': 'pbsrollout',
    'version': '0.2.1',
    'description': '',
    'long_description': None,
    'author': 'dimitri',
    'author_email': 'dimitriwyzlic@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
